

#include <wchar.h>
#include "Parser.h"
#include "Scanner.h"


namespace JSON {


void Parser::SynErr(int n) {
	if (errDist >= minErrDist) errors->SynErr(la->line, la->col, n);
	errDist = 0;
}

void Parser::SemErr(const wchar_t* msg) {
	if (errDist >= minErrDist) errors->Error(t->line, t->col, msg);
	errDist = 0;
}

void Parser::Get() {
	for (;;) {
		t = la;
		la = scanner->Scan();
		if (la->kind <= maxT) { ++errDist; break; }

		if (dummyToken != t) {
			dummyToken->kind = t->kind;
			dummyToken->pos = t->pos;
			dummyToken->col = t->col;
			dummyToken->line = t->line;
			dummyToken->next = NULL;
			coco_string_delete(dummyToken->val);
			dummyToken->val = coco_string_create(t->val);
			t = dummyToken;
		}
		la = t;
	}
}

void Parser::Expect(int n) {
	if (la->kind==n) Get(); else { SynErr(n); }
}

void Parser::ExpectWeak(int n, int follow) {
	if (la->kind == n) Get();
	else {
		SynErr(n);
		while (!StartOf(follow)) Get();
	}
}

bool Parser::WeakSeparator(int n, int syFol, int repFol) {
	if (la->kind == n) {Get(); return true;}
	else if (StartOf(repFol)) {return false;}
	else {
		SynErr(n);
		while (!(StartOf(syFol) || StartOf(repFol) || StartOf(0))) {
			Get();
		}
		return StartOf(syFol);
	}
}

void Parser::JSON() {
		if (la->kind == 3) {
			object();
		} else if (la->kind == 7) {
			array();
		} else SynErr(13);
}

void Parser::object() {
		Expect(3);
		if (la->kind == 1) {
			pair();
			while (la->kind == 4) {
				Get();
				pair();
			}
		}
		Expect(5);
}

void Parser::array() {
		Expect(7);
		if (StartOf(1)) {
			value();
			while (la->kind == 4) {
				Get();
				value();
			}
		}
		Expect(8);
}

void Parser::pair() {
		Expect(1);
		key = JSON_string(pool, t);        
		Expect(6);
		value();
		ap_table_setn(tabl, key, val);     
}

void Parser::value() {
		switch (la->kind) {
		case 1: {
			Get();
			val = JSON_string(pool, t);        
			break;
		}
		case 2: {
			Get();
			val = JSON_string(pool, t);        
			break;
		}
		case 3: {
			object();
			break;
		}
		case 7: {
			array();
			break;
		}
		case 9: {
			Get();
			val = JSON_string(pool, t);        
			break;
		}
		case 10: {
			Get();
			val = JSON_string(pool, t);        
			break;
		}
		case 11: {
			Get();
			val = 0;                           
			break;
		}
		default: SynErr(14); break;
		}
}



void Parser::Parse() {
	t = NULL;
	la = dummyToken = new Token();
	la->val = coco_string_create(L"Dummy Token");
	Get();
	JSON();

	Expect(0);
}

Parser::Parser(Scanner *scanner) {
	maxT = 12;

	minErrDist = 2;
	errDist = minErrDist;
	this->scanner = scanner;
	errors = new Errors();
   	dummyToken = NULL;
   	t = la = NULL;
}

bool Parser::StartOf(int s) {
	const bool T = true;
	const bool x = false;

	static bool set[2][14] = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x},
		{x,T,T,T, x,x,x,T, x,T,T,T, x,x}
	};



	return set[s][la->kind];
}

Parser::~Parser() {
	delete errors;
	delete dummyToken;
}

Errors::Errors() {
	count = 0;
}

void Errors::SynErr(int line, int col, int n) {
        // In mod_ndb we do not remark upon client's JSON errors
        count++;
        return;
        
	wchar_t* s;
	switch (n) {
			case 0: s = coco_string_create(L"EOF expected"); break;
			case 1: s = coco_string_create(L"string expected"); break;
			case 2: s = coco_string_create(L"number expected"); break;
			case 3: s = coco_string_create(L"\"{\" expected"); break;
			case 4: s = coco_string_create(L"\",\" expected"); break;
			case 5: s = coco_string_create(L"\"}\" expected"); break;
			case 6: s = coco_string_create(L"\":\" expected"); break;
			case 7: s = coco_string_create(L"\"[\" expected"); break;
			case 8: s = coco_string_create(L"\"]\" expected"); break;
			case 9: s = coco_string_create(L"\"true\" expected"); break;
			case 10: s = coco_string_create(L"\"false\" expected"); break;
			case 11: s = coco_string_create(L"\"null\" expected"); break;
			case 12: s = coco_string_create(L"??? expected"); break;
			case 13: s = coco_string_create(L"invalid JSON"); break;
			case 14: s = coco_string_create(L"invalid value"); break;

		default:
		{
			wchar_t format[20];
			coco_swprintf(format, 20, L"error %d", n);
			s = coco_string_create(format);
		}
		break;
	}
	wprintf(L"-- line %d col %d: %ls\n", line, col, s);
	coco_string_delete(s);
	count++;
}

void Errors::Error(int line, int col, const wchar_t *s) {
	count++;
}

void Errors::Warning(int line, int col, const wchar_t *s) {
}

void Errors::Warning(const wchar_t *s) {
}

void Errors::Exception(const wchar_t* s) {
	wprintf(L"%ls", s); 
	exit(1);
}

} // namespace


